
import numpy as np

from . import qfn

### Defining some traditional gates
#: 2x2 identity gate :math:`\begin{bmatrix} 1 & 0 \\ 0 & 1 \end{bmatrix}`
IDEN = np.array([
    [1,0],
    [0,1]])
#: 2x2 logical NOT gate :math:`\begin{bmatrix} 0 & 1 \\ 1 & 0 \end{bmatrix}`
NOT = np.array([
    [0,1],
    [1,0]])
#: 2x4 logical OR gate :math:`\begin{bmatrix} 1 & 0 & 0 & 0 \\ 0 & 1 & 1 & 1 \end{bmatrix}`
OR = np.array([
    [1,0,0,0],
    [0,1,1,1]])
#: 2x4 logical AND gate :math:`\begin{bmatrix} 1 & 1 & 1 & 0 \\ 0 & 0 & 0 & 1 \end{bmatrix}`
AND = np.array([
    [1,1,1,0],
    [0,0,0,1]])
#: 2x4 logical XOR gate :math:`\begin{bmatrix} 1 & 0 & 0 & 1 \\ 0 & 1 & 1 & 0 \end{bmatrix}`
XOR = np.array([
    [1,0,0,1],
    [0,1,1,0]])
#: 2x4 logical NOR gate :math:`\begin{bmatrix} 0 & 1 & 1 & 1 \\ 1 & 0 & 0 & 0 \end{bmatrix}`
NOR  = NOT @ OR
#: 2x4 logical NAND gate :math:`\begin{bmatrix} 0 & 0 & 0 & 1 \\ 1 & 1 & 1 & 0 \end{bmatrix}`
NAND = NOT @ AND
#: 4x2 COPY gate :math:`\begin{bmatrix} 1 & 0 \\ 0 & 0 \\ 0 & 0 \\ 0 & 1 \end{bmatrix}`
COPY = np.array([
    [1,0],
    [0,0],
    [0,0],
    [0,1]])
#: 4x4 SWAP gate :math:`\begin{bmatrix} 1 & 0 & 0 & 0 \\ 0 & 0 & 1 & 0 \\ 0 & 1 & 0 & 0 \\ 0 & 0 & 0 & 1 \end{bmatrix}`
SWAP = np.array([
    [1,0,0,0],
    [0,0,1,0],
    [0,1,0,0],
    [0,0,0,1]])


### Defining some quantum gates
#: 4x4 Quantum CNOT gate: :math:`\begin{bmatrix} 1 & 0 & 0 & 0 \\ 0 & 1 & 0 & 0 \\ 0 & 0 & 0 & 1 \\ 0 & 0 & 1 & 0 \end{bmatrix}`
CNOT = np.kron(IDEN,XOR) @ np.kron(COPY,IDEN)
#: 8x8 Quantum TOFFOLI gate: :math:`\begin{bmatrix} 1 & 0 & 0 & 0 & 0 & 0 & 0 & 0 \\ 0 & 1 & 0 & 0 & 0 & 0 & 0 & 0 \\ 0 & 0 & 1 & 0 & 0 & 0 & 0 & 0 \\ 0 & 0 & 0 & 1 & 0 & 0 & 0 & 0 \\ 0 & 0 & 0 & 0 & 1 & 0 & 0 & 0 \\ 0 & 0 & 0 & 0 & 0 & 1 & 0 & 0 \\ 0 & 0 & 0 & 0 & 0 & 0 & 0 & 1 \\ 0 & 0 & 0 & 0 & 0 & 0 & 1 & 0 \end{bmatrix}`
TOFFOLI = (
    np.kron(np.kron(IDEN,IDEN),XOR) @
    np.kron(np.kron(np.kron(IDEN,IDEN),AND),IDEN) @
    np.kron(np.kron(np.kron(IDEN,SWAP),IDEN),IDEN) @
    np.kron(np.kron(COPY,COPY),IDEN)
    )
#: 2x2 Quantum HADAMARD gate :math:`\begin{bmatrix} \frac{1}{\sqrt{2}} && \frac{1}{\sqrt{2}} \\ \frac{1}{\sqrt{2}} && \frac{-1}{\sqrt{2}} \end{bmatrix}`
HAD = np.array([[1,1],[1,-1]]) * np.exp2(-1/2)



"""qgates.qfn

Contains helper functions for working with
quantum gates.
"""

from typing import Union, Iterable

from functools import reduce as _reduce
from operator import matmul as _matmul

import numpy as np

from . import states
from . import gates

__version__ = "1.0.3"


def tens(a: Union[int,float,list,np.ndarray],*b: Union[int,float,list,np.ndarray]) -> np.ndarray:
    """
    Computes the tensor product of
    two (or more) vectors/matrices.

    :param a: Start value for tensor product
    :param b: One or more other values for tensor product
    """
    return _reduce(np.kron,b,a)

def matmul(a: np.ndarray,*b: np.ndarray) -> np.ndarray:
    """
    Performs matrix multiplication between
    two or more

    :param a: Start value for matrix multiplication
    :param b: One or more other values for tensor product
    """
    return _reduce(_matmul,b,a)

def state(arr: Union[int,Iterable]) -> np.ndarray:
    """Creates a state vector
    from either an iterable of
    1s and 0s (or from an int that
    will be converted to a state
    vector based on its binary
    representation.

    :param arr: Int or iterable for creating state vector
    """
    if isinstance(arr,int):
        arr = [int(b) for b in bin(arr)[2:]]
    s = states.QB1 if arr[0] else states.QB0
    for n in arr[1:]:
        if n == 0:
            s = tens(s,states.QB0)
        else:
            s = tens(s,states.QB1)
    return s


def conjugate(v: Union[complex,list,np.ndarray],*args,**kwargs) -> Union[complex,np.ndarray]:
    """
    Returns the complex conjugate of v.

    :param v: number, vector, or matrix to conjugate
    """
    return np.conjugate(v,*args,**kwargs)




import numpy as np

from . import qfn

#### Basic Qubit States #####
#: Qubit with the state: :math:`\ket{0}`
QB0  = np.array([1,0])
#: Qubit with the state: :math:`\ket{1}`
QB1  = np.array([0,1])
#: Qubit with the state: :math:`\ket{00}`
QB00 = np.kron(QB0, QB0)
#: Qubit with the state: :math:`\ket{01}`
QB01 = np.kron(QB0, QB1)
#: Qubit with the state: :math:`\ket{10}`
QB10 = np.kron(QB1, QB0)
#: Qubit with the state: :math:`\ket{11}`
QB11 = np.kron(QB1, QB1)


##### Bell States #####
#: Bell state 00 :math:`\ket{\Phi^+} = \frac{1}{\sqrt2}(\ket{0_A} \otimes \ket{0_B} + \ket{1_A} \otimes \ket{1_B})`
BELL00 = 2**(-1/2)*qfn.state((0,0)) + 2**(-1/2)*qfn.state((1,1))
#: Bell state 00 :math:`\ket{\Phi^-} = \frac{1}{\sqrt2}(\ket{0_A} \otimes \ket{0_B} - \ket{1_A} \otimes \ket{1_B})`
BELL01 = 2**(-1/2)*qfn.state((0,1)) + 2**(-1/2)*qfn.state((1,0))
#: Bell state 00 :math:`\ket{\Psi^+} = \frac{1}{\sqrt2}(\ket{0_A} \otimes \ket{1_B} + \ket{1_A} \otimes \ket{0_B})`
BELL10 = 2**(-1/2)*qfn.state((0,0)) - 2**(-1/2)*qfn.state((1,1))
#: Bell state 00 :math:`\ket{\Psi^+} = \frac{1}{\sqrt2}(\ket{0_A} \otimes \ket{1_B} - \ket{1_A} \otimes \ket{0_B})`
BELL11 = 2**(-1/2)*qfn.state((0,1)) - 2**(-1/2)*qfn.state((1,0))
